/*
 * ===========
 * = shvec.c =
 * ===========
 *
 * history:
 * ========
 * $Id: shvec.c,v 1.1 1999/11/02 17:25:35 fv Exp $
 *
 * history:
 * ========
 * $Id: shvec.c,v 1.1 1999/11/02 17:25:35 fv Exp $
 *
 * description:
 * ============
 * The function "computeShortVectors" computes short vectors of a lattice L
 * which is given by a Gram-matrix. There are 4 different modes.
 * - (SHVEC_MODE_BOUND)
 *   compute all vectors of L which length is bounded by a given constant
 * - (SHVEC_MODE_SHORTEST_VECTORS)
 *   compute the minimal vectors of L
 * - (SHVEC_MODE_MINIMUM)
 *   compute only the minimum of L
 * - (SHVEC_MODE_THETA_SERIES)
 *   count only the number of vectors of L which length is bounded by a
 *   given constant
 * Additionally there are many options.
 * 
 * The algorithm used is based on the algorithm of Fincke and Pohst.
 * (see Henri Cohen --  A Course in Computational Algebraic Number Theory)
 *
 * E-Mail:
 * =======
 * frank.vallentin@math.uni-dortmund.de
 */


#include "shvec.h"

#include "la_support.h"
#include "lll_basis.h"
#include "panic.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>



/*
 * ======================
 * = constants & macros =
 * ======================
 */

#define SHVEC_MAX 1000 /* allocate always SHVEC_MAX vectors */
#define STOP_COMPUTATION 69
#define ROUND(x) (int) floor(x + 0.5)
#define SHVEC_TEST_EPSILON 0.01


/*
 * =================
 * = debug options =
 * =================
 */


/*
#define SHVEC_DEBUG
#define SHVEC_computeIt_DEBUG
#define SHVEC_initShvecReq_DEBUG
#define SHVEC_destroyShvecReq_DEBUG
#define SHVEC_initShvecInfo_DEBUG
#define SHVEC_destroyShvecInfo_DEBUG
#define SHVEC_computeShvec_DEBUG
#define SHVEC_insertBound_DEBUG
#define SHVEC_CHECK
*/



#ifdef SHVEC_DEBUG

static void printInternals(internals ints)
{
  printf("(\n");
  printf("dim: %d\n", ints.dim);
  printf("gram_matrix: "); 
  if (ints.gram_matrix == NULL)
    printf("NULL\n");
  else {
    printf("\n");
    doublePrintGramMatrix(ints.dim, ints.gram_matrix);
  }
  printf("integral: %d\n", ints.integral);
  printf("cholesky_decomposition: %x\n", (int) ints.cholesky_decomposition);
  printf("quadratic_form: ");
  if (ints.quadratic_form == NULL)
    printf("NULL\n");
  else {
    printf("\n");
    doublePrintSquareMatrix(ints.dim, ints.quadratic_form);
  }
  printf("transformation: ");
  if (ints.transformation == NULL)
    printf("NULL\n");
  else {
    intPrintSquareMatrix(ints.dim, ints.transformation);
  }
  printf("sigma: ");
  if (ints.sigma == NULL)
    printf("NULL\n");
  else {
    printf("\n");
    intPrintVector(ints.dim, ints.sigma);
  }
  printf("dummy_vector: ");
  if (ints.sigma == NULL)
    printf("NULL\n");
  else {
    printf("\n");
    intPrintVector(ints.dim, ints.dummy_vector);
  }
  printf("even: %d\n", ints.even);
  printf("check: %d\n", ints.check);
  printf(")\n");
}


static void printShvecReq(shvec_request request)
{

  printf("(\n");
  printf("mode: ");
  if (request->mode == SHVEC_MODE_BOUND)
    printf("SHVEC_MODE_BOUND\n");
  else if (request->mode == SHVEC_MODE_SHORTEST_VECTORS)
    printf("SHVEC_MODE_SHORTEST_VECTORS\n");
  else if (request->mode == SHVEC_MODE_MINIMUM)
    printf("SHVEC_MODE_MINIMUM\n");
  else if (request->mode == SHVEC_MODE_THETA_SERIES)
    printf("SHVEC_MODE_THETA_SERIES\n");
  else printf("wrong mode\n");  
  printf("bound: %f\n", request->bound);
  printf("number: %d\n", request->number);
  printf("coset: ");
  if (request->coset == NULL)
    printf("NULL\n");
  else {
    doublePrintVector(request->ints.dim, request->coset);
  }
  printf("ints: "); printInternals(request->ints);
  printf(")\n");
}


static void printShvecInfo(shvec_info info)
{
  int
    i,
    j;

  printf("(\n");
  printf("request: ");
  printShvecReq(&(info->request));
  printf("short_vectors_count: %d\n", info->short_vectors_count);
  printf("short_vectors_number: %d\n", info->short_vectors_number);
  printf("short_vectors: ");
  if (info->short_vectors != NULL) {
    printf("\n");
    for (i = 0; i < info->short_vectors_number; i++) {
      printf("[%4d]: ", i);
      intPrintVector(info->request.ints.dim, info->short_vectors[i]);
    }
  }
  else printf("NULL\n");
  printf("short_vectors_norm: ");
  if (info->short_vectors_norm != NULL) {
    printf("\n");
    for (i = 0; i < info->short_vectors_number; i++)
      printf("[%4d]: %f\n", i, info->short_vectors_norm[i]);
  }
  else printf("NULL\n");
  printf("minimum: %f\n", info->minimum);
  printf("theta_series_bound: %f\n", info->theta_series_bound);
  printf("theta_series: ");
  if (info->theta_series != NULL) {
    printf("\n");
    for (i = 0; i <= info->theta_series_bound; i++) {
      printf("[%2d]: %d\n", i, info->theta_series[i]);
    }
  }
  else printf("NULL\n");
  printf("shells: ");
  if (info->shells != NULL) {
    printf("\n");
    for (i = 0; i <= info->theta_series_bound; i++) {
      printf("[%2d]: ", i);
      for (j = 0; j < info->theta_series[i] / 2; j++)
	printf(" %d", info->shells[i][j]);
      printf("\n");
    }
  }
  else printf("NULL\n");
  printf(")\n");  
}

#endif



/*
 * ======================
 * = internal functions =
 * ======================
 */



/*
 * internal memory management
 * ==========================
 * SHVEC_malloc, SHVEC_realloc, SHVEC_free
 */

static void *SHVEC_malloc(size_t size)
{
  void
    *ptr;

  ptr = (void *) calloc(1, size);
  if (ptr == NULL) {
    PANIC;
  }
  return ptr;
}

static void *SHVEC_realloc(void *ptr, size_t size)
{
  void
    *new_ptr;
  
  new_ptr = (void *) realloc(ptr, size);
  if (new_ptr == NULL) {
    PANIC;
  }
  return new_ptr;
}

static void SHVEC_free(void *ptr)
{
  if (ptr == NULL) {
    PANIC;
  }
  free(ptr);
}



/*
 * ============================
 * = functions to add vectors =
 * ============================
 */



/*
 * insertBound
 * ===========
 * don't declare static
 */

int insertBound(shvec_info info,
		int *vector,
		int *c,
		int coset,
		double norm)
{
  int
    *sigma,
    number,
    i,
    vector_norm,
    shell_bound,
    dim;
#ifdef SHVEC_CHECK
  double
    vector_norm_check,
    *check_vec;
#endif

#ifdef SHVEC_insertBound_DEBUG
  printf("(insertBound begin\n");
  printf("==================\n");
  printf("info: "); printShvecInfo(info);
  printf("vector: "); intPrintVector(info->request.ints.dim, vector);
  printf("c: "); intPrintVector(info->request.ints.dim, c);
  printf("coset: %d\n", coset);
  printf("norm: %f\n", norm);
#endif

  dim = info->request.ints.dim;
  number = info->short_vectors_number;

  if (info->request.ints.integral && !coset) {
    vector_norm = ROUND(norm);
    norm = (int) vector_norm;
  }

  sigma = info->request.ints.sigma;

  /*
   * Arrays evtl. vergr"ossern
   */

  if (number % SHVEC_MAX == 0) {
    info->short_vectors =
      SHVEC_realloc(info->short_vectors,
		    SHVEC_MAX * (number/SHVEC_MAX + 1) * sizeof(int *));
    info->short_vectors[number] =
      SHVEC_malloc(dim * sizeof(int) * SHVEC_MAX);
    for (i = 1; i < SHVEC_MAX; i++) {
      info->short_vectors[i + number] =
	info->short_vectors[number] + i * dim;
    }
    info->short_vectors_norm =
      SHVEC_realloc(info->short_vectors_norm,
		    SHVEC_MAX * (number/SHVEC_MAX + 1) * sizeof(double));
  }

  /* 
   * theta_series und shells nur fuer ganzzahlige Gitter ohne Nebenklassen
   */

  if (info->request.ints.integral && !coset) {

    if (info->theta_series_bound < vector_norm) {
      info->theta_series =
	SHVEC_realloc(info->theta_series, (vector_norm + 1) * sizeof(int));
      info->shells =
	SHVEC_realloc(info->shells, (vector_norm + 1) * sizeof(int *));
      for (i = info->theta_series_bound + 1; i <= vector_norm; i++) {
      info->theta_series[i] = 0;
      info->shells[i] = NULL;
      }
      info->theta_series_bound = vector_norm;
    }
    
    shell_bound = info->theta_series[vector_norm]/2;
    if (shell_bound % SHVEC_MAX == 0) {
      info->shells[vector_norm] =
	SHVEC_realloc(info->shells[vector_norm],
		      SHVEC_MAX * (shell_bound / SHVEC_MAX + 1) * sizeof(int));
    }
  }

  /*
   * Werte eintragen
   */

  for (i = 0; i < dim; i++)
    (info->request.ints.dummy_vector)[i] = vector[sigma[i]] - c[sigma[i]];

  intMultSquareMatrixVector(dim,
			    info->request.ints.transformation,
			    info->request.ints.dummy_vector,
			    info->short_vectors[number]);

  /*
   * Nebenklasse anpassen
   */

  info->short_vectors_norm[number] = norm;
  
  if (info->request.ints.integral && !coset) {
    info->theta_series[vector_norm] = info->theta_series[vector_norm] + 2;
    info->shells[vector_norm][shell_bound] = number;
  }

  /*
   * Zeiger hochz"ahlen
   */

  info->short_vectors_count++;
  info->short_vectors_number++;

  /*
   * neues Minimum gefunden?
   */
  
  if (norm < info->minimum || info->minimum == -1.0)
    info->minimum = norm;

  /*
   * Vektornorm evtl. abchecken
   */

#ifdef SHVEC_CHECK
    doubleMakeVector(dim, &check_vec);
    for (i = 0; i < dim; i++)
      check_vec[i] = (double) info->short_vectors[number][i]
	+ info->request.coset[i];
    doubleComputeVectorNorm(dim, info->request.ints.gram_matrix, check_vec,
			    &vector_norm_check);
    if (fabs(norm - vector_norm_check) >= SHVEC_TEST_EPSILON)
      fprintf(stderr, "shvec.c (insertBound) wrong vector_norm!\n");
    doubleDestroyVector(dim, &check_vec);
#endif

  /*
   * Schranke beruecksichtigen
   */

  if (info->short_vectors_number == info->request.number)
    return STOP_COMPUTATION;
  else
    return 0;

#ifdef SHVEC_insertBound_DEBUG
  printf("insertBound end\n");
  printf("===============\n");
  printf("info: "); printShvecInfo(info);
  printf("vector: "); intPrintVector(info->request.ints.dim, vector);
  printf("c: "); intPrintVector(info->request.ints.dim, c);
  printf("coset: %d\n", coset);
  printf("norm: %f\n", norm);
  printf(")\n");
#endif
}



/*
 * insertThetaSeries
 * =================
 * don't declare static
 */

int insertThetaSeries(shvec_info info,
		      int *vector,
		      int *c,
		      int coset,
		      double norm)
{
  int
    vector_norm,
    i;

  info->short_vectors_count++;

  vector_norm = ROUND(norm);

  info->short_vectors_number++;

  if (info->theta_series_bound < (double) vector_norm) {
    info->theta_series =
      SHVEC_realloc(info->theta_series,
		    (vector_norm + 1) * sizeof(int));
    for (i = info->theta_series_bound + 1; i <= vector_norm; i++)
      info->theta_series[i] = 0;
    info->theta_series_bound = (double) vector_norm;
  }

  info->theta_series[vector_norm] = info->theta_series[vector_norm] + 2;

  if (norm < info->minimum || info->minimum == -1.0) {
    info->minimum = norm;
  }
  
  return 0;
}



/*
 * insertStop
 * ==========
 * don't declare static
 */

int insertStop(shvec_info info,
		      int *vector,
		      int *c,
		      int coset,
		      double norm)
{
  info->minimum = norm;
  return STOP_COMPUTATION;
}



/*
 * =================================
 * = algorithm of Fincke and Pohst =
 * =================================
 */

int computeIt(shvec_info info,
	      int (*insert)())
{
  int
    sigma,
    coset,
    result,
    not_finished,
    needs_new_bound,
    *x,
    *c,
    i,
    j,
    dim;
  double
    *double_vec,
    *C,
    sum,
    *X,
    bound,
    *T,
    *U,
    *L,
    **q,
    Z;

#ifdef SHVEC_computeIt_DEBUG
  printf("(computeIt begin\n");
  printf("================\n");
  printf("info: "); printShvecInfo(info);
  printf("insert: %x\n", (int) insert); 
#endif

  /*
   * initialization
   */

  result = 0;
  dim = info->request.ints.dim;
  bound = info->request.bound;

  doubleMakeVector(dim, &T);
  doubleMakeVector(dim, &U);
  doubleMakeVector(dim, &L);
  doubleMakeVector(dim, &X);
  doubleMakeVector(dim, &C);
  doubleMakeVector(dim, &double_vec);
  intMakeVector(dim, &x);
  intMakeVector(dim, &c);
  
  coset = 0;
  i = 0;
  while (i < dim && !coset) {
    coset = (info->request.coset[i] != 0.0);
    i++;
  }
  
  /* 
   * Normalize the entries of the coset-vectors to N = [-0.5, 0.5),
   * then info->request.coset[i] = C[i] + c[i] and C[i] lies in N.
   * Take care of sigma.
   */
    
  doubleMultSquareMatrixVector(dim, info->request.ints.inv_transformation,
			       info->request.coset, double_vec);
  
  for (i = 0; i < dim; i++) {
    sigma = info->request.ints.sigma[i];
    c[sigma] = (int) double_vec[i];
    C[sigma] = double_vec[i] - c[sigma];
    if (C[sigma] >= 0.5) {
      C[sigma] = C[sigma] - 1.0;
      c[sigma]++;
    }
    else if (C[sigma] < -0.5) {
      C[sigma] = C[sigma] + 1.0;
      c[sigma]--;
    }
  }

  /*
   * Computation!
   */

  not_finished = 1;
  needs_new_bound = 1;
  i = dim - 1;
  T[i] = bound;
  U[i] = 0.0;
  q = info->request.ints.quadratic_form;
  
  while (not_finished) {

    if (needs_new_bound) {
      Z = sqrt((T[i] + SHVEC_ELLIPSOID_EPSILON) / q[i][i]);
      L[i] = floor(Z - U[i] - C[i]);
      X[i] = ceil(-Z - U[i] - C[i]) - 1.0;
      x[i] = (int) X[i];
      needs_new_bound = 0;
    }

    x[i]++;
    X[i] = X[i] + 1.0;
    
    if (X[i] <= L[i]) {
	
      if (i == 0) {

	if (!coset) {
	  j = dim - 1;
	  not_finished = 0;
	  while (j >= 0 && !not_finished) {
	    not_finished = (x[j] != 0);
	    j--;
	  }
	  if (!not_finished) goto urgent_stop;
	}
	
	result = insert(info, x, c, coset,
			bound - T[0] + q[0][0] * (X[0] + C[0] + U[0]) *
			(X[0] + C[0] + U[0]));
	if (result == STOP_COMPUTATION) goto urgent_stop;
	
      }
      else {
	i--;
	sum = 0.0;
	for (j = i + 1; j < dim; j++)
	  sum += q[i][j] * (X[j] + C[j]);
	U[i] = sum;
	T[i] = 
	  T[i+1] - q[i+1][i+1] * (X[i+1] + C[i+1] + U[i+1]) * (X[i+1] + C[i+1] + U[i+1]);
	needs_new_bound = 1;
      }
    }
    else {
      i++;
      if (i >= dim) not_finished = 0;
    }
  } /* while (not_finished) */
  
urgent_stop:
    
  /*
   * free memory
   */

  doubleDestroyVector(dim, &T);
  doubleDestroyVector(dim, &U);
  doubleDestroyVector(dim, &L);
  doubleDestroyVector(dim, &X);
  doubleDestroyVector(dim, &C);
  doubleDestroyVector(dim, &double_vec);
  intDestroyVector(dim, &c);
  intDestroyVector(dim, &x);

#ifdef SHVEC_computeIt_DEBUG
  printf("computeIt end\n");
  printf("=============\n");
  printf("info: "); printShvecInfo(info);
  printf("insert: %x\n", (int) insert); 
  printf(")\n");
#endif

  return result;
}


/*
 * computeMinimum
 * ==============
 */

static int computeMinimum(shvec_info info)
{
  int
    result,
    finished,
    dim,
    coset,
    i;
  double
    *double_vec,
    *C,
    step_size;

  dim = info->request.ints.dim;

  doubleMakeVector(dim, &C);
  doubleMakeVector(dim, &double_vec);

  coset = 0;
  i = 0;
  while (i < dim && !coset) {
    coset = (info->request.coset[i] != 0.0);
    i++;
  }


  /* 
   * Normalize the entries of the coset-vectors to N = [-0.5, 0.5),
   * then info->request.coset[i] = C[i] + c[i] and C[i] lies in N.
   */

  for (i = 0; i < dim; i++) {
    C[i] = info->request.coset[i] - (int) info->request.coset[i];
    if (C[i] >= 0.5)
      C[i] = C[i] - 1.0;
    else if (C[i] < -0.5)
      C[i] = C[i] + 1.0;
  }

  /*
   * compute minimum
   */

  if (coset)
    doubleComputeVectorNorm(dim, info->request.ints.gram_matrix,
			    C, &(info->minimum));
  else {
    info->minimum = info->request.ints.gram_matrix[0][0];
    for (i = 1; i < dim; i++)
      if (info->minimum > info->request.ints.gram_matrix[i][i])
	info->minimum = info->request.ints.gram_matrix[i][i];
  }

  doubleDestroyVector(dim, &C);
  doubleDestroyVector(dim, &double_vec);
  
  if (coset || !info->request.ints.integral) step_size = SHVEC_STEP_EPSILON;
  else if (info->request.ints.even) step_size = 2.0;
  else step_size = 1.0;

  finished = 0;
  while (!finished) {
    info->request.bound = info->minimum - step_size;
    result = computeIt(info, insertStop);
    if (result != STOP_COMPUTATION) finished = 1;
  }  
  info->minimum = info->request.bound + step_size;
  
  return 0;
}



/*
 * copyShvecReq
 * ============
 * copy a busy request to an empty request
 */

static void copyShvecReq(shvec_request req_from,
			 shvec_request req_to)
{
  int
    dim;
  
  dim = req_from->ints.dim;

  req_to->mode = req_from->mode;
  req_to->bound = req_from->bound;
  req_to->number = req_from->number;
  doubleMakeVector(dim, &(req_to->coset));
  doubleCopyVector(dim, req_from->coset, req_to->coset);
  req_to->ints.dim = dim;
  doubleMakeSquareMatrix(dim, &(req_to->ints.gram_matrix));
  doubleCopySquareMatrix(dim, req_from->ints.gram_matrix, 
			 req_to->ints.gram_matrix);
  req_to->ints.integral = req_from->ints.integral;
  req_to->ints.cholesky_decomposition = req_from->ints.cholesky_decomposition;
  doubleMakeSquareMatrix(dim, &(req_to->ints.quadratic_form));
  doubleCopySquareMatrix(dim, req_from->ints.quadratic_form,
			 req_to->ints.quadratic_form);
  intMakeSquareMatrix(dim, &(req_to->ints.transformation));
  intCopySquareMatrix(dim, req_from->ints.transformation,
		      req_to->ints.transformation);
  intMakeVector(dim, &(req_to->ints.sigma));
  doubleMakeSquareMatrix(dim, &(req_to->ints.inv_transformation));
  doubleCopySquareMatrix(dim, req_from->ints.inv_transformation,
			 req_to->ints.inv_transformation);
  intCopyVector(dim, req_from->ints.sigma, req_to->ints.sigma);
  intMakeVector(dim, &(req_to->ints.dummy_vector));
  req_to->ints.even = req_from->ints.even;
  req_to->ints.check = req_from->ints.check;
}  



/*
 * ======================
 * = exported functions =
 * ======================
 */


/*
 * initShvecReq
 * ============
 */

void initShvecReq(int dim,
		  double **gram_matrix,
		  int (*cholesky_decomposition)(),
		  int check,
		  shvec_request request)
{
  double
    **up_trig_mat,
    **inv_up_trig_mat,
    **double_transformation,
    **reduced_mat,
    **inv_reduced_mat,
    **q,
    *row_norm,
    temp;
  int
    success,
    i,
    j,
    t;

#ifdef SHVEC_initShvecReq_DEBUG
  printf("(initShvecReq begin\n");
  printf("-------------------\n");
  printf("dim: %d\n", dim);
  printf("gram_matrix:\n"); doublePrintGramMatrix(dim, gram_matrix);
  printf("cholesky_decomposition: %x\n", (int) cholesky_decomposition);
  printf("check: %d\n", check);
  printf("request: assumed to be undefined\n"); 
#endif

  if (dim < 2 ||
      gram_matrix == NULL) {
    fprintf(stderr, "shvec.c (initShvecReq) wrong input!\n");
    abort();
  }

  doubleMakeVector(dim, &(request->coset));

  request->ints.dim = dim;
  request->ints.check = check;

  /*
   * test: integral Gram-matrix?
   *       YES -> test: even Gram-matrix?
   */

  request->ints.integral = 1;
  i = 0; j = 0;
  while (i < dim && request->ints.integral) {
    while (j < dim && request->ints.integral) {
      request->ints.integral = 
	(gram_matrix[i][j] - (int) gram_matrix[i][j] == 0.0);
      j++;
    }
    i++;
  }

  if (request->ints.integral) {
    request->ints.even = 1;
    for (i = 0; i < dim; i++)
      if (ROUND(gram_matrix[i][i]) % 2 != 0)
	request->ints.even = 0;
  }
  else
    request->ints.even = 0;

  
  /* 
   * copy Gram-Matrix to shvec_request and test for eveness
   */

  doubleMakeSquareMatrix(dim, &(request->ints.gram_matrix));
  doubleCopySquareMatrix(dim, gram_matrix, request->ints.gram_matrix);

  
  /*
   * compute the other fields
   */

  request->mode = 0;
  request->bound = 0.0;
  request->number = 0;

  if (cholesky_decomposition == NULL)
    request->ints.cholesky_decomposition = cholesky;
  else
    request->ints.cholesky_decomposition = cholesky_decomposition;

  
  /*
   * ... some precomputations ...
   */

  doubleMakeSquareMatrix(dim, &(request->ints.quadratic_form));
  intMakeSquareMatrix(dim, &(request->ints.transformation));
  doubleMakeSquareMatrix(dim, &(request->ints.inv_transformation));
  intMakeVector(dim, &(request->ints.sigma));
  intMakeVector(dim, &(request->ints.dummy_vector));

  doubleMakeSquareMatrix(dim, &up_trig_mat);
  doubleMakeSquareMatrix(dim, &inv_up_trig_mat);
  doubleMakeSquareMatrix(dim, &double_transformation);
  doubleMakeSquareMatrix(dim, &reduced_mat);
  doubleMakeSquareMatrix(dim, &inv_reduced_mat);
  doubleMakeVector(dim, &row_norm);
  
  success = request->ints.cholesky_decomposition(dim, gram_matrix,
						 up_trig_mat);
  if (!success) {
    fprintf(stderr, "shvec.c (initShvecReq) cholesky went wrong!\n");
    abort();
  }

  inverseUpperMatrix(dim, up_trig_mat, inv_up_trig_mat);
  computeLLLReducedBasis(dim, inv_up_trig_mat, inv_reduced_mat, check);
  inverseMatrix(dim, inv_reduced_mat, reduced_mat);

  doubleSquareMatrixMult(dim, inv_up_trig_mat, reduced_mat, 
			 double_transformation);
  double2intSquareMatrix(dim, double_transformation,
			 request->ints.transformation);
  inverseMatrix(dim, double_transformation, request->ints.inv_transformation);

  /*
   * sort rows and columns
   */
  
  for (i = 0; i < dim; i++) {
    rowNorm(dim, inv_reduced_mat, i, &(row_norm[i]));
    request->ints.sigma[i] = i;
  }

  for (i = 0; i < dim; i++)
    for (j = i + 1; j < dim; j++) {
      if (row_norm[i] < row_norm[j]) {
	t = request->ints.sigma[i];
	request->ints.sigma[i] = request->ints.sigma[j];
	request->ints.sigma[j] = t;
	temp = row_norm[i];
	row_norm[i] = row_norm[j];
	row_norm[j] = temp;
      }
    }

  /* DIRTY: use up_trig_mat as temporary space */

  doubleCopySquareMatrix(dim, reduced_mat, up_trig_mat);
  for (i = 0; i < dim; i++)
    for (j = 0; j < dim; j++)
      reduced_mat[j][request->ints.sigma[i]] = up_trig_mat[j][i];

  trMatTimesMat(dim, reduced_mat, up_trig_mat);

  q = request->ints.quadratic_form;

  success = request->ints.cholesky_decomposition(dim, up_trig_mat, q);

  if (!success) {
    fprintf(stderr, "shvec.c (initShvecReq) cholesky went wrong!\n");
    abort();
  }
  
  for (i = 0; i < dim; i++)
    q[i][i] = q[i][i] * q[i][i];

  for (i = 0; i < dim; i++)
    for (j = i + 1; j < dim; j++)
      q[i][j] = q[i][j] / sqrt(q[i][i]);

  for (i = 0; i < dim; i++)
    for (j = 0; j < i; j++)
      q[i][j] = q[j][i];

  doubleDestroyVector(dim, &row_norm);
  doubleDestroySquareMatrix(dim, &up_trig_mat);
  doubleDestroySquareMatrix(dim, &inv_up_trig_mat);
  doubleDestroySquareMatrix(dim, &double_transformation);
  doubleDestroySquareMatrix(dim, &reduced_mat);
  doubleDestroySquareMatrix(dim, &inv_reduced_mat);

#ifdef SHVEC_initShvecReq_DEBUG
  printf("initShvecReq end\n");
  printf("----------------\n");
  printf("dim: %d\n", dim);
  printf("gram_matrix:\n"); doublePrintGramMatrix(dim, gram_matrix);
  printf("cholesky_decomposition: %x\n", (int) cholesky_decomposition);
  printf("check: %d\n", check);
  printf("request: "); printShvecReq(request);
  printf(")\n");
#endif
}



/*
 * destroyShvecReq
 * ===============
 */

void destroyShvecReq(shvec_request request)
{
  int
    dim;

  if (request == NULL) {
    fprintf
      (stderr,
       "shvec.c (destroyShvecReq) wrong input!\n");
    abort();
  }

#ifdef SHVEC_destroyShvecReq_DEBUG
  printf("(destroyShvecReq begin\n");
  printf("----------------------\n");
  printf("request: "); printShvecReq(request);
#endif

  dim = request->ints.dim;

  request->mode = 0;
  request->bound = 0.0;
  request->number = 0;
  if (request->coset != NULL)
    doubleDestroyVector(dim, &(request->coset));
  request->ints.cholesky_decomposition = NULL;

  request->ints.dim = 0;
  if (request->ints.gram_matrix != NULL)
    doubleDestroySquareMatrix(dim, &(request->ints.gram_matrix));
  request->ints.integral = 0;
  if (request->ints.quadratic_form != NULL)
    doubleDestroySquareMatrix(dim, &(request->ints.quadratic_form));
  if (request->ints.transformation != NULL)
    intDestroySquareMatrix(dim, &(request->ints.transformation));
  if (request->ints.inv_transformation != NULL)
    doubleDestroySquareMatrix(dim, &(request->ints.inv_transformation));
  if (request->ints.sigma != NULL)
    intDestroyVector(dim, &(request->ints.sigma));
  if (request->ints.dummy_vector != NULL)
    intDestroyVector(dim, &(request->ints.dummy_vector));
  request->ints.even = 0;
  request->ints.check = 0;

#ifdef SHVEC_destroyShvecReq_DEBUG
  printf("destroyShvecReq end\n");
  printf("-------------------\n");
  printf("request: "); printShvecReq(request);
  printf(")\n");
#endif
}



/*
 * initShvecInfo
 * =============
 */

void initShvecInfo(shvec_info info)
{

#ifdef SHVEC_initShvecInfo_DEBUG
  printf("(initShvecInfo begin\n");
  printf("--------------------\n");
  printf("info: assumed to be undefinded\n");
#endif

  info->request.mode = 0;
  info->request.bound = 0.0;
  info->request.number = 0;
  info->request.coset = NULL;

  info->request.ints.dim = 0;
  info->request.ints.gram_matrix = NULL;
  info->request.ints.integral = 0;
  info->request.ints.cholesky_decomposition = NULL;
  info->request.ints.quadratic_form = NULL;
  info->request.ints.transformation = NULL;
  info->request.ints.inv_transformation = NULL;
  info->request.ints.sigma = NULL;
  info->request.ints.dummy_vector = NULL;
  info->request.ints.even = 0;
  info->request.ints.check = 0;

  info->short_vectors_count = 0;
  info->short_vectors_number = 0;
  info->short_vectors = NULL;
  info->short_vectors_norm = NULL;
  info->minimum = -1.0;
  info->theta_series_bound = 0.0;
  info->theta_series = NULL;
  info->shells = NULL;
 
#ifdef SHVEC_initShvecInfo_DEBUG
  printf("initShvecInfo end\n");
  printf("-----------------\n");
  printf("info: "); printShvecInfo(info);
  printf(")\n");
#endif
}



/*
 * destroyShvecInfo
 * ================
 */

void destroyShvecInfo(shvec_info info)
{
  int
    i;

#ifdef SHVEC_destroyShvecInfo_DEBUG
  printf("(destroyShvecInfo begin\n");
  printf("-----------------------\n");
  printf("info: "); printShvecInfo(info);
#endif
  
  if (info == NULL) {
    fprintf
      (stderr,
       "shvec.c (destroyShvecInfo) wrong input!");
    abort();
  }

  info->short_vectors_count = 0;

  if (info->short_vectors != NULL) {
    for (i = 0; i < info->short_vectors_number / SHVEC_MAX + 1; i++)
      SHVEC_free(info->short_vectors[i * SHVEC_MAX]);
    SHVEC_free(info->short_vectors);
  }
  info->short_vectors = NULL;

  if (info->short_vectors_norm != NULL)
    SHVEC_free(info->short_vectors_norm);
  info->short_vectors_norm = NULL;
  
  info->short_vectors_number = 0;
  info->minimum = -1.0;

  if (info->theta_series != NULL)
    SHVEC_free(info->theta_series);
  info->theta_series = NULL;

  if (info->shells != NULL) {
    for (i = 0; i <= info->theta_series_bound; i++)
      if (info->shells[i] != NULL)
	SHVEC_free(info->shells[i]);
    SHVEC_free(info->shells);
  }
  info->shells = NULL;
  
  info->theta_series_bound = 0.0;

  destroyShvecReq(&info->request);

#ifdef SHVEC_destroyShvecInfo_DEBUG
  printf("destroyShvecInfo end\n");
  printf("--------------------\n");
  printf("info: "); printShvecInfo(info);
  printf(")\n");
#endif
}



/*
 * computeShvec
 * ============
 */

void computeShvec(shvec_request request,
		  shvec_info info)
{
  int
    dim,
    result,
    coset,
    i;

#ifdef SHVEC_computeShvec_DEBUG
  printf("(computeShvec begin\n");
  printf("-------------------\n");
  printf("request: "); printShvecReq(request);
  printf("info: "); printShvecInfo(info);
#endif

  if (request == NULL ||
      info == NULL) {
    fprintf(stderr, "shvec.c (computeShvec) wrong input!");
    abort();
  }

  destroyShvecInfo(info);
  
  /*
   * test the request
   */

  dim = (request->ints).dim;

  coset = 0;
  i = 0;
  while (i < dim && !coset) {
    coset = (request->coset[i] != 0.0);
    i++;
  }
  
  switch (request->mode) {
  case SHVEC_MODE_BOUND:
    if (request->bound <= 0.0 || request->number < 0)
      goto fatal_request_error;
    break;
  case SHVEC_MODE_SHORTEST_VECTORS:
    if (request->bound != 0.0 || request->number < 0)
      goto fatal_request_error;
    break;
  case SHVEC_MODE_MINIMUM:
    if (request->bound != 0.0 || request->number != 0)
      goto fatal_request_error;
    break;
  case SHVEC_MODE_THETA_SERIES:
    if (request->bound <= 0.0 || request->number != 0 || 
	request->ints.integral == 0 || coset)
      goto fatal_request_error;
    break;
  default:
    goto fatal_request_error;
  }

  /*
   * copy request
   */
  
  copyShvecReq(request, &(info->request));

  /*
   * choose mode
   */

  /* SHVEC_BOUND */

  if (request->mode == SHVEC_MODE_BOUND)
    result = computeIt(info, insertBound);

  /* SHVEC_MODE_SHORTEST_VECTORS */

  else if (request->mode == SHVEC_MODE_SHORTEST_VECTORS) {
    result = computeMinimum(info);
    info->request.bound = info->minimum;
    result = computeIt(info, insertBound);
  }

  /* SHVEC_MODE_MINIMUM */
  
  else if (request->mode == SHVEC_MODE_MINIMUM) {
    result = computeMinimum(info);
  }

  /* SHVEC_MODE_THETA_SERIES */

  else if (request->mode == SHVEC_MODE_THETA_SERIES)
    result = computeIt(info, insertThetaSeries);

  /*
   * 0-th coefficient of theta-series is always 1
   */
  
  if (info->theta_series != NULL) info->theta_series[0] = 1;
  if (info->shells != NULL) info->shells[0] = NULL;

#ifdef SHVEC_computeShvec_DEBUG
  printf("computeShvec end\n");
  printf("----------------\n");
  printf("info: "); printShvecInfo(info);
  printf("request: "); printShvecReq(request);
  printf(")\n");
#endif

  return;

fatal_request_error:
  fprintf(stderr, "shvec.c (computeShvec): wrong options!\n");
  abort();
}
