/*
 * ===========
 * = theta.c =
 * ===========
 *
 * history:
 * ========
 * $Id: theta.c,v 1.1 1999/11/02 17:25:35 fv Exp $
 *
 * description:
 * ============
 * computes coefficients of degree 2 thetaseries
 *
 * e-mail:
 * =======
 * frank.vallentin@math.uni-dortmund.de
 */

#include <ctype.h>
#include <malloc.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "shvec.h"
#include "la_support.h"


/*
 * computeThetaCoeff
 * =================
 */

void computeThetaCoeff(int dim,
		       double **gram_matrix,
		       shvec_info info,
		       int **siegel_matrix,
		       int *coeff)
{
  int
    **gram,
    a,
    b,
    c,
    i,
    j,
    prod,
    *vec;
  
  intMakeSquareMatrix(dim, &gram);
  for (i = 0; i < dim; i++)
    for (j = 0; j < dim; j++)
      gram[i][j] = (int) gram_matrix[i][j];

  *coeff = 0;

  a = siegel_matrix[0][0];
  b = siegel_matrix[0][1];
  c = siegel_matrix[1][1];

  intMakeVector(dim, &vec);

  for (i = 0; i < info->theta_series[a]/2; i++) {
    intMultSquareMatrixVector(dim, gram,
			       info->short_vectors[info->shells[a][i]], vec);
    for (j = 0; j < info->theta_series[c]/2; j++) {
      intStandardScalarProduct(dim, vec,
				info->short_vectors[info->shells[c][j]], &prod);
      if (prod == b || -prod == b) {
	(*coeff)++;
	if (prod == 0) (*coeff)++;
      }
    }
  }

  *coeff = *coeff * 2;

  intDestroyVector(dim, &vec);
  intDestroySquareMatrix(dim, &gram);
}



/*
 * is_prime
 * ========
 * using trial division
 */

int is_prime(int x)
{
  int i;
  double w;
  
  if (x == 2) return 1;
  if (x % 2 == 0) return 0;

  w = ceil(sqrt(x));
  for (i = 3; i < w; i = i + 2)
    if (x % i == 0) return 0;
  return 1;
}


/*
 * factorize
 * =========
 */

void factorize(int x)
{
  int
    i,
    exponent,
    y;

  y = x;
  for (i = 2; i < (int) ceil(sqrt(x)) && y != 1; i++) {
    if (is_prime(i)) {
      exponent = 0;
      while (y % i == 0) {
	y = y / i;
	exponent++;
      }
      if (exponent == 1) {
	printf(" %d", i);
      }
      else if (exponent > 1) {
	printf(" %d^%d", i, exponent);
      }
    }
  }
  printf(".\n");
}


/*
 * main
 * ====
 */

int main(int argc,
	 char *argv[])
{ 
  int
    check,
    gram,
    even,
    bound,
    a,
    b,
    c,
    max,
    coeff,
    i,
    j,
    dim,
    siegel_dim,
    **siegel_matrix;
  double
    **gram_matrix;
  shvec_request
    req;
  shvec_info
    info;

  req = (shvec_request) malloc(sizeof(struct shvec_request_struct));
  if (req == NULL) {
    fprintf(stderr, "theta.c: not enough memory!\n");
    exit(1);
  }

  info = (shvec_info) malloc(sizeof(struct shvec_info_struct));
  if (info == NULL) {
    fprintf(stderr, "theta.c: not enough memory!\n");
    exit(1);
  }

  check = 1;
  gram = 0;
  bound = 0;
  siegel_dim = 2;
  even = 1;

  /* 
   * user input
   */

  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-' && argv[i][1] == 'g')
      gram = 1;
    if (argv[i][0] == '-' && argv[i][1] == 'b') {
      if (i + 1 < argc)
	sscanf(argv[i + 1], "%d", &bound);
      if (i + 1 >= argc || bound <= 0) {
	fprintf(stderr, "bound is not properly set!\n");
	exit(1);
      }	
    }
    if (argv[i][0] == '-' && argv[i][1] == 'h') {
      printf("usage: theta <in >out [-g] [-b y] [-h]\n");
      printf("in: read dim and gram-matrix from in\n");
      printf("out: write output to out\n");
      printf("-g: print gram-matrix\n");
      printf("-b y: bound\n");
      printf("-h: print this help\n");
      return 0;
    }
  }

  /*
   * read gram-matrix
   */

  scanf("%d", &dim);

  doubleMakeSquareMatrix(dim, &gram_matrix);
  
  for (i = 0; i < dim; i++)
    for (j = 0; j <= i; j++) {
      scanf("%lf", &gram_matrix[i][j]);
      gram_matrix[j][i] = gram_matrix[i][j];
    }

  if (gram) {
    printf("%d\n", dim);
    doublePrintGramMatrix(dim, gram_matrix);
  }

  for (i = 0; i < dim; i++)
    if ((int) gram_matrix[i][i] % 2 != 0) even = 0;
    
  /*
   * if bound == 0, read another matrix
   */

  intMakeSquareMatrix(siegel_dim, &siegel_matrix);

  if (bound == 0) {
    for (i = 0; i < siegel_dim; i++)
      for (j = 0; j <= i; j++) {
	scanf("%d", &siegel_matrix[i][j]);
	siegel_matrix[j][i] = siegel_matrix[i][j];
      }
    if (gram) {
      printf("%d\n", siegel_dim);
      intPrintGramMatrix(siegel_dim, siegel_matrix);
    }
    
  }

  /*
   * computation
   */

  if (bound == 0)
    {
      max = siegel_matrix[0][0];
      for (i = 1; i < siegel_dim; i++)
	if (siegel_matrix[i][i] > max) max = siegel_matrix[i][i];
    
      initShvecReq(dim, gram_matrix, NULL, check, req);
      req->mode = SHVEC_MODE_BOUND;
      req->bound = max;
      req->number = 0;
      initShvecInfo(info);
      computeShvec(req, info);

      printf("M = [%d, %d; %d, %d], det(M) = %d, a_L(M) = ",
	     siegel_matrix[0][0], siegel_matrix[0][1],
	     siegel_matrix[1][0], siegel_matrix[1][1], 
	     siegel_matrix[0][0]*siegel_matrix[1][1] -
	     siegel_matrix[0][1]*siegel_matrix[0][1]
	     );
      fflush(stdout);

      computeThetaCoeff(dim, gram_matrix, info, siegel_matrix, &coeff);

      printf("%d =", coeff);
      factorize(coeff);
  }
  else
    {
      max = 0;
      for (c = 1; c <= bound; c++)
	for (a = 1; a <= c; a++)
	  for (b = 0; 2*b <= a && a*c - b*b <= bound; b++) {
	    if (!(even && (a % 2 != 0 || c % 2 != 0))) {
	      if (c > max) max = c;
	    }
	  }

      initShvecReq(dim, gram_matrix, NULL, check, req);
      req->mode = SHVEC_MODE_BOUND;
      req->bound = max;
      req->number = 0;
      initShvecInfo(info);
      computeShvec(req, info);

      for (c = 1; c <= bound; c++)
	for (a = 1; a <= c; a++)
	  for (b = 0; 2*b <= a && a*c - b*b <= bound; b++) {
	    if (!(even && (a % 2 != 0 || c % 2 != 0))) {
	      siegel_matrix[0][0] = a;
	      siegel_matrix[0][1] = b;
	      siegel_matrix[1][0] = b;
	      siegel_matrix[1][1] = c;

	      printf("M = [%d, %d; %d, %d], det(M) = %d, a_L(M) = ",
		     siegel_matrix[0][0], siegel_matrix[0][1],
		     siegel_matrix[1][0], siegel_matrix[1][1], 
		     siegel_matrix[0][0]*siegel_matrix[1][1] -
		     siegel_matrix[0][1]*siegel_matrix[0][1]
		     );
	      fflush(stdout);

	      computeThetaCoeff(dim, gram_matrix, info, siegel_matrix, &coeff);

	      printf("%d =", coeff);
	      factorize(coeff);
	    }
	  }
    }

  /*
   * clean up
   */

  destroyShvecInfo(info);
  destroyShvecReq(req);
  doubleDestroySquareMatrix(dim, &gram_matrix);
  intDestroySquareMatrix(siegel_dim, &siegel_matrix);
  free(info);
  free(req);

  return 0;
}
