/*
 * ==========
 * = voro.c =
 * ==========
 *
 * history:
 * ========
 * $Id: voro.c,v 1.1 1999/11/02 17:25:35 fv Exp $
 *
 * description:
 * ============
 * compute the facets of a Voronoi-cell of a lattice
 * output is input for polymake
 * (original version is written by Ralf Gerkmann)
 *
 * e-mail
 * ======
 * ralf.gerkmann@math.uni-dortmund.de
 * frank.vallentin@math.uni-dortmund.de
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include "shvec.h"
#include "la_support.h"

#define TRUE 1
#define FALSE 0
#define MQ_EPSILON 0.00001
/*
#define DEBUG
*/

void print_help(void)
{
  printf("usage: voro <in >out [-g] [-h]\n");
  printf("in: read dim and gram-matrix from in\n");
  printf("out: write polymake output to out\n");
  printf("-g: print gram-matrix\n");
  printf("-h: print this help\n"); 
}


void read_invar(int *dim, double ***mat)
{
  int i,j;
  int input;

  scanf("%d\n", dim);
  *mat = (double**) malloc(sizeof(double*)*(*dim));
  if ((*mat)==NULL)
    {
      fprintf(stderr, "voro: Not enough memory\n");
    };
  for (i=0; i<*dim; i++)
    {
      (*mat)[i]= (double*) malloc(sizeof(double)*(*dim));
      if ((*mat)[i]==NULL)
	{
	  fprintf(stderr, "voro: Not enough memory\n");
	};
    };
  for (i=0; i<*dim; i++)
    {
      for (j=0; j<=i; j++)
	{
	  scanf("%d", &input);
          (*mat)[i][j] = (double)input;
	  (*mat)[j][i]= (*mat)[i][j];
	};
    }
}

void trash_mat(int dim, double **mat)
{
  int i;

  for (i=0; i<dim; i++)
    {
      free(mat[i]);
    };
  free(mat);
}

void computeOutput(int dim, shvec_request request, shvec_info info, double **mat)
{
  double **basis, **vorvects;
  long i,j,k;
  long vors, mask, cur_bit, maxsize;
  double distance;  

  /* initialization */

  vors = 0;
  maxsize = 1;
  maxsize = maxsize << dim;
  basis = (double**) malloc(sizeof(double*)*dim);
  vorvects = (double**) malloc(sizeof(double*)*(maxsize));
  for (i=0; i<dim; i++)
    {
      basis[i]= (double*) malloc(sizeof(double)*dim);
    };
  for (i=0; i<maxsize; i++)
    {
      vorvects[i]= (double*) malloc(sizeof(double)*dim);
    };

  /* computation of the basis */

  cholesky(dim, mat, basis);

#ifdef DEBUG
  printf("basis of cholesky decomposition:\n");
  for (i=0; i<dim; i++)
    {
      for (j=0; j<dim; j++)
	printf("%10f ", basis[j][i]);
      printf("\n");
    };
  printf("\n");
  printf("Maxsize = %d\n", maxsize);
#endif

  for (i=0; i<dim; i++)
    for (j=0; j<dim; j++)
      mat[i][j]= 4*mat[i][j];
  initShvecReq(dim, mat, NULL, 0, request);
  initShvecInfo(info);

  /* computation of the coefficient of the shortest coset vectors */

  for (i=1; i<maxsize; i++)
    {
      request->mode= SHVEC_MODE_SHORTEST_VECTORS;
      request->number = 0;
      mask = 1;
      for (cur_bit=0; cur_bit<dim; cur_bit++)
	{
	  if ((mask & i) != 0)
	    {
	      request->coset[cur_bit]=0.5;
	    }
	  else
	    {
	      request->coset[cur_bit]=0.0;
	    };
	  mask = mask*2;
	};

#ifdef DEBUG
      printf("coset vectors: ");
      for (j=0; j<dim; j++)
	printf("%10f ",request->coset[j]);
      printf("\n");
#endif

      /*      destroyShvecInfo(info);
	      initShvecInfo(info);
      */
      
      computeShvec(request, info);

#ifdef DEBUG
      printf("shortest vectors:\n");
      for (k=0; k<info->short_vectors_number;k++)
	{
	  for (j=0; j<dim; j++)
	    printf("%d ", info->short_vectors[k][j]);
	  printf("\n");
	};
      printf("\n");
#endif

      /* computation of the voronoi vectors */

      if (info->short_vectors_number==2)
	{
	  for (j=0; j<dim; j++)
	    {
	      vorvects[vors][j]=0.0;
	      for (k=0; k<dim; k++)
		{
		  vorvects[vors][j]+=2*(info->short_vectors[0][k]
				      +request->coset[k])*basis[j][k];
		};
	    };
	  vors++;
	};
    };
  
  /* output as polymake input */

  printf("AMBIENT_DIM\n%d\n\n", dim);
  printf("DIM\n%d\n\n", dim);

  printf("FACETS(float)\n");

  for (i=0; i<vors; i++) {
    distance=0;
    for (j=0; j<dim; j++)
      distance+=vorvects[i][j]*vorvects[i][j];
    printf("%10.20f ", distance);

    for (j=0; j<dim; j++)
      printf("%10.20f ", (-1)*vorvects[i][j]);
    printf("\n");  
  }

  for (i=0; i<vors; i++) {
    distance=0;
    for (j=0; j<dim; j++)
      distance+=vorvects[i][j]*vorvects[i][j];
    printf("%10.20f ", distance);
  
  for (j=0; j<dim; j++)
      printf("%10.20f ", vorvects[i][j]);
    printf("\n");
  }

  for (i=0; i<dim; i++)
    {
      free(basis[i]);
    };
  for (i=0; i<maxsize; i++)
    {
      free(vorvects[i]);
    };
  free(vorvects);
  free(basis);
}

int main(int argc, char **argv)
{
  double **mat;
  int dim, i, j;
  shvec_request request;
  shvec_info info;

  if ((argv[1][0]=='-') && (argv[1][1]=='h'))
    {
      print_help();
      return 0;
    };

  read_invar(&dim, &mat);

  if ((argv[1][0]=='-') && (argv[1][1]=='g'))
    {
      for (i=0; i<dim; i++)
	{
	  for (j=0; j<=i; j++)
	    {
	      printf("%f ", mat[i][j]);
	    };
	  printf("\n");
	};
    }
  
  request = (shvec_request) malloc(sizeof(struct shvec_request_struct));
  if (request==NULL)
    {
      fprintf(stderr, "voro: Not enough memory\n");
    };
  info = (shvec_info) malloc(sizeof(struct shvec_info_struct));
  if (info==NULL)
    {
      fprintf(stderr, "voro: Not enough memory\n");
    };
  
  computeOutput(dim, request, info, mat);
  
  destroyShvecReq(request);
  destroyShvecInfo(info);
  
  trash_mat(dim, mat);

  return 0;
}

